// SPDX-License-Identifier: GPL-2.0
/*
 * Example program to show how to use UIO to map a PCIe bar
 * Copyright (C) 2021, Xilinx, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
/*
 * Copyright 2021 Xilinx, Inc
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/*
 * UIO PAN card driver, heavily based on the following driver in the Linux source tree:
 *      drivers/uio/uio_cif.c
 *
 * Module parameter manipulation code, as illustrated in
 *      https://devarea.com/linux-kernel-development-kernel-module-parameters/
 *
 * Centos packages needed:
 *     kernel-devel
 *     elfutils-libelf-devel
 */

#include <linux/device.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/slab.h>
#include <linux/uio_driver.h>

#include <asm/io.h>

/*
 * This is the BAR used for communication with the R5
 */
#define COMM_BAR    4

static int pan_uio_pci_probe(struct pci_dev *dev,
					const struct pci_device_id *id)
{
	struct uio_info *info;

	info = kzalloc(sizeof(struct uio_info), GFP_KERNEL);
	if (!info)
		return -ENOMEM;

	if (pci_enable_device(dev))
		goto out_free;

	if (pci_request_regions(dev, "pan_uio"))
		goto out_disable;

	info->mem[0].addr = pci_resource_start(dev, COMM_BAR);
	if (!info->mem[0].addr)
		goto out_release;

	info->mem[0].size = pci_resource_len(dev, COMM_BAR);
	info->mem[0].memtype = UIO_MEM_PHYS;
    info->name = "pan_uio";
	info->version = "0.0.1";

	if (uio_register_device(&dev->dev, info))
		goto out_unmap;

	pci_set_drvdata(dev, info);

	return 0;
out_unmap:
out_release:
	pci_release_regions(dev);
out_disable:
	pci_disable_device(dev);
out_free:
	kfree (info);
	return -ENODEV;
}

static void pan_uio_pci_remove(struct pci_dev *dev)
{
	struct uio_info *info = pci_get_drvdata(dev);

	uio_unregister_device(info);
	pci_release_regions(dev);
	pci_disable_device(dev);

	kfree (info);
}

static struct pci_device_id pan_uio_pci_ids[] = {
	{
		.vendor =	PCI_VENDOR_ID_XILINX,
		.device =	0xb03f,
		.subvendor =	PCI_VENDOR_ID_XILINX,
		.subdevice =	0x0007,
	},
	{ 0, }
};

static struct pci_driver pan_uio_pci_driver = {
	.name = "pan_uio",
	.id_table = pan_uio_pci_ids,
	.probe = pan_uio_pci_probe,
	.remove = pan_uio_pci_remove,
};

module_pci_driver(pan_uio_pci_driver);
MODULE_DEVICE_TABLE(pci, pan_uio_pci_ids);
MODULE_LICENSE("GPL v2");
MODULE_AUTHOR("Xilinx, Inc.");
