/*
 * Copyright 2021 Xilinx, Inc
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include <x86_app.h>

static bool GetFlashInfo(
        Sys* pSys,
        uint32_t qLoc,
        ImageLoc* pImageLoc
        )
{
    BoardInfo* pBI;

    pImageLoc->qSubsectorNumber = 0;
    pImageLoc->qPEBStart = 0;
    pImageLoc->qPEBEnd = 0;
    pImageLoc->qRBEIndex = 0;

    pBI = pSys->pBI;
    if(!pBI) {
        Warn(pSys,0,"No board selected");
        return false;
    }

    switch(tolower((char)qLoc)) {
    default:
        Warn(pSys,0,"Unknown multiboot character '%c'",(char)qLoc);
        return false;

    case 'a':
        pImageLoc->qPEBStart = 0;
        pImageLoc->qPEBEnd = pBI->qPEBOffsetB - 1;
        pImageLoc->qRBEIndex = 0;
        break;

    case 'b':
        pImageLoc->qPEBStart = pBI->qPEBOffsetB;
        pImageLoc->qPEBEnd = pBI->qPEBOffsetG - 1;
        pImageLoc->qRBEIndex = 1;
        break;

    case 'g':
        pImageLoc->qPEBStart = pBI->qPEBOffsetG;
        pImageLoc->qPEBEnd = pBI->qNumPEBs - 2;
        pImageLoc->qRBEIndex = 2;
        break;
    }
    pImageLoc->qSubsectorNumber =
        ((pBI->qPEBSize/pBI->qSubsectorSize) * (pBI->qNumPEBs - 1))
        +
        pImageLoc->qRBEIndex
        ;
    return true;
}

/*
 * Erase a location's PEBs and subsector
 */
bool EraseLoc(Sys* pSys,uint32_t qLoc)
{
    ImageLoc tIL;
    uint32_t qRet;
    uint32_t qErrBSP;
    char pcErrString[ERR_STRING_BUFSIZE];

    if(!GetFlashInfo(pSys,qLoc,&tIL)) {
        return false;
    }

    qRet = ErasePEBs(pSys,tIL.qPEBStart,tIL.qPEBEnd,&qErrBSP);
    if(ERR_NONE != (ERR)qRet) {
        Warn(pSys,0,"ErasePEBs() -> %s (ErrBSP=%u)",GetErrString(qRet,pcErrString),qErrBSP);
        return false;
    }
    memset((void*)pSys->ppRBE[tIL.qRBEIndex],0xff,pSys->pBI->qSubsectorSize);
    qRet = EraseSubsector(pSys,tIL.qSubsectorNumber,&qErrBSP);
    if(ERR_NONE != (ERR)qRet) {
        Warn(pSys,0,"EraseSubsector() -> %s (ErrBSP=%u)",GetErrString(qRet,pcErrString),qErrBSP);
        return false;
    }

    return true;
}

static bool WriteRBEToFlash(Sys* pSys,ReservedBlockEntry* pRBE,uint32_t qSubsectorNumber)
{
    uint32_t qRet;
    uint32_t qErrBSP;
    uint32_t qErrCodeSave;
    uint32_t qErrBSPSave;
    char pcErrString[ERR_STRING_BUFSIZE];

    qErrCodeSave = pRBE->qErrCode;
    pRBE->qErrCode = 0;
    qErrBSPSave = pRBE->qErrBSP;
    pRBE->qErrBSP = 0;

    pRBE->qEntryChecksum = Crc32_ComputeBuf( 
                               0,                               /* uint32_t inCrc32, */
                               (void*)pRBE,                     /* const void *buf, */
                               sizeof(ReservedBlockEntry) - 4   /* uint32_t bufLen */
                               );
    pRBE->qVersion = htole32(pRBE->qVersion);
    pRBE->qImageOffset = htole32(pRBE->qImageOffset);
    pRBE->qImageSize = htole32(pRBE->qImageSize);
    pRBE->qImageChecksum = htole32(pRBE->qImageChecksum);
    pRBE->qEntryChecksum = htole32(pRBE->qEntryChecksum);
    qRet = WriteSubsector(pSys,qSubsectorNumber,(const uint8_t*)pRBE,&qErrBSP);

    pRBE->qErrCode = qErrCodeSave;
    pRBE->qErrBSP = qErrBSPSave;
    pRBE->qVersion = le32toh(pRBE->qVersion);
    pRBE->qImageOffset = le32toh(pRBE->qImageOffset);
    pRBE->qImageSize = le32toh(pRBE->qImageSize);
    pRBE->qImageChecksum = le32toh(pRBE->qImageChecksum);
    pRBE->qEntryChecksum = le32toh(pRBE->qEntryChecksum);

    if(ERR_NONE != (ERR)qRet) {
        Warn(pSys,0,"WriteSubsector() -> %s (ErrBSP=%u)",GetErrString(qRet,pcErrString),qErrBSP);
        return false;
    }
    return true;
}

/*
 * Program image at previously erased PEBs and subsectors
 */
bool ProgramImageAtLoc(
        Sys* pSys,
        ImageInfo* pII,
        uint32_t qLoc               /* One of 'a', 'b', 'g' */
        )
{
    ImageLoc tIL;
    uint32_t qRet;
    uint32_t qPEBNumber;
    BoardInfo* pBI;
    uint8_t* pbSrc;
    ReservedBlockEntry* pRBE;
    uint32_t qRemainingBytes;
    uint32_t qBytesToWrite;
    uint8_t* pbLastSrcChunk;
    uint32_t qErrBSP;
    char pcErrString[ERR_STRING_BUFSIZE];

    pBI = pSys->pBI;

    if(!GetFlashInfo(pSys,qLoc,&tIL)) {
        return false;
    }

    for(
                qRet = ERR_NONE,
                    qPEBNumber = tIL.qPEBStart,
                    pbSrc = (uint8_t*)pII->pxMap,
                    pbLastSrcChunk = 0,
                    qRemainingBytes = pII->qSize;

                (ERR_NONE == (ERR)qRet) && (qRemainingBytes > 0);

                qPEBNumber++,
                    pbSrc += qBytesToWrite,
                    qRemainingBytes -= qBytesToWrite
            ) {
        if(qRemainingBytes >= pBI->qPEBSize) {
            qBytesToWrite = pBI->qPEBSize;
        } else {
            qBytesToWrite = qRemainingBytes;
            pbLastSrcChunk = malloc(pBI->qPEBSize);
            if(!pbLastSrcChunk) {
                Warn(pSys,errno,"LastSrcChunk");
                return false;
            }
            memcpy(pbLastSrcChunk,pbSrc,qRemainingBytes);
            memset(pbLastSrcChunk+qRemainingBytes,0xff,pBI->qPEBSize - qRemainingBytes);
            pbSrc = pbLastSrcChunk;
        }
        qRet = WritePEB(pSys,qPEBNumber,pbSrc,&qErrBSP);
    }
    free(pbLastSrcChunk);
    if(ERR_NONE != (ERR)qRet) {
        Warn(pSys,0,"WritePEB() -> %s (ErrBSP=%u)",GetErrString(qRet,pcErrString),qErrBSP);
        return false;
    }

    pRBE = pSys->ppRBE[tIL.qRBEIndex];
    pRBE->qVersion = RBE_VERSION;
    pRBE->qImageOffset = tIL.qPEBStart * pBI->qPEBSize;
    pRBE->qImageSize = pII->qSize;
    pRBE->qImageChecksum = pII->qChecksum;
    memcpy(pRBE->pbImageId,pII->tCR.ppcTokens[2],II_ID_BYTES);
    memset(pRBE->pbReserved,0xff,RBE_RESERVED_BYTES);
    pRBE->qErrCode = ERR_NONE;
    pRBE->qErrBSP = 0;
    return WriteRBEToFlash(pSys,pRBE,tIL.qSubsectorNumber);
}

bool CopyAtoB(Sys* pSys)
{
    ImageLoc tILA;
    ImageLoc tILB;
    uint32_t qRet;
    uint32_t qPEBNumberA;
    uint32_t qPEBNumberB;
    uint32_t qNumPEBs;
    ReservedBlockEntry* pRBE_A;
    ReservedBlockEntry* pRBE_B;
    uint32_t qErrBSP;
    char pcErrString[ERR_STRING_BUFSIZE];

    if(!GetFlashInfo(pSys,'a',&tILA)) {
        return false;
    }
    if(!GetFlashInfo(pSys,'b',&tILB)) {
        return false;
    }
    for(
            qPEBNumberA = tILA.qPEBStart,
                qPEBNumberB = tILB.qPEBStart,
                qNumPEBs = tILA.qPEBEnd - tILA.qPEBStart + 1
                ;
            qNumPEBs > 0;
            qNumPEBs--
            ) {
        qRet = ReadPEBInternal(pSys,qPEBNumberA++,&qErrBSP);
        if(ERR_NONE != (ERR)qRet) {
            Warn(pSys,0,"ReadPEBInternal() -> %s (ErrBSP=%u)",GetErrString(qRet,pcErrString),qErrBSP);
            break;
        }
        qRet = WritePEBInternal(pSys,qPEBNumberB++,&qErrBSP);
        if(ERR_NONE != (ERR)qRet) {
            Warn(pSys,0,"WritePEBInternal() -> %s (ErrBSP=%u)",GetErrString(qRet,pcErrString),qErrBSP);
            break;
        }
    }

    pRBE_A = pSys->ppRBE[0];
    pRBE_B = pSys->ppRBE[1];
    *pRBE_B = *pRBE_A;
    pRBE_B->qImageOffset = tILB.qPEBStart * pSys->pBI->qPEBSize;
    return WriteRBEToFlash(pSys,pRBE_B,tILB.qSubsectorNumber);
}
