/*
 * Copyright 2021 Xilinx, Inc
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include <x86_app.h>

/*
 * --------------------------------------------------------------
 * List manipulation
 * --------------------------------------------------------------
 */

/*
 * --------------------------------------------------------------
 * List Manipulation
 * List header is an extension of ListElem as well
 *     so empty list is header pointing to itself
 * An empty list can be initialized multiple times
 * --------------------------------------------------------------
 */
void InitListHead(ListHead* pListHead,Sys* pSys,const char* pcFile)
{
    pListHead->tElem.pNext = pListHead->tElem.pPrev = &pListHead->tElem;
    pListHead->pSys = pSys;
    pListHead->pcFile = pcFile;
}

/*
 * Unsorted quick append at end
 */
void InsertListElem(ListHead* pListHead,ListElem* pElem)
{
    pElem->pNext = &pListHead->tElem;
    pElem->pPrev = pListHead->tElem.pPrev;

    pElem->pPrev->pNext = pElem;
    pListHead->tElem.pPrev = pElem;
}

/*
 * Iterate through list. Run CheckElem on each list element in turn using pxArg.
 * If any call to CheckElem returns true,return it
 * Otherwise, return 0 at end.
 *
 * List element is not dereferenced after calling CheckElem.
 */
ListElem* CheckElems(
        ListHead* pListHead,
        bool (*rCheckElem)(ListHead* pListHead,ListElem* pListElem,void* pxArg),
        void* pxArg
        )
{
    ListElem* pLE;
    ListElem* pNext;

    for(pLE = pListHead->tElem.pNext; pLE != &pListHead->tElem; pLE = pNext) {
        pNext = pLE->pNext;
        if(rCheckElem(pListHead,pLE,pxArg)) {
            return pLE;
        }
    }
    return (ListElem*)0;
}

/*
 * CheckElem compatible
 *
 * This wrapper can be used to free an entire list
 */
static bool FreeListElem(ListHead* pListHead,ListElem* pListElem,void* pxArg)
{
    free((void*)pListElem);
    return false;
}

void FreeList(ListHead* pListHead)
{
    (void)CheckElems(pListHead,FreeListElem,(void*)0);
    InitListHead(pListHead,pListHead->pSys,pListHead->pcFile);
}

/*
 * --------------------------------------------------------------
 * --------------------------------------------------------------
 */
bool GetBaseNumber(
        Sys* pSys,
        const char* pcFile,
        int iLine,
        const char* pcVariable,
        char* pcToken,
        int iBase,
        uint32_t* pqVal
        )
{
    char* endptr;

    errno = 0;
    *pqVal = strtoul(pcToken,&endptr,iBase);
    if(errno || *endptr) {
        Die(pSys,0,"%s:%d: %s: Invalid base-%d number '%s'",
                pcFile,
                iLine,
                pcVariable,
                iBase,
                pcToken
                );
        return false;
    }
    return true;
}

/*
 * --------------------------------------------------------------
 * --------------------------------------------------------------
 */
void PrettyPrint(Sys* pSys,uint32_t qCount,uint8_t* pbBuf)
{
    uint32_t qOffset;

    for(qOffset = 0; qOffset < 16; qOffset++) {
    }
}

/*
 * --------------------------------------------------------------
 * Break buffer into a max# of tokens, with user-specified
 * separators.
 * Return number of tokens found
 * --------------------------------------------------------------
 */
int Tokenize(
        Sys* pSys,
        const char* pcFile,
        int iLine,
        char* pcBuf,
        const char* pcSep,
        int iMaxTokens,
        char* ppcTokens[]
        )
{
    char* pcToken;
    char* saveptr;
    char* pc;
    int iNumTokens;

    for(iNumTokens = 0; iNumTokens <= iMaxTokens; ) {
        pcToken = strtok_r((0 == iNumTokens)?pcBuf:0,pcSep,&saveptr);
        if(!pcToken) {
            break;
        }
        if(*pcToken == '#') {
            break;
        }
        pc = 0;
        if((pc = strchr(pcToken,'#'))) {
            *pc = 0;
        }
        if(iNumTokens == iMaxTokens) {
            Warn(pSys,0,"%s:%d: Too many tokens: max %d",pcFile,iLine,iMaxTokens);
            return -1;
        }
        ppcTokens[iNumTokens++] = pcToken;
        if(pc) {
            break;
        }
    }
    ppcTokens[iNumTokens] = 0;
    return iNumTokens;
}

/*
 * Map a file so we can memcpy to/from directly to it
 */
bool MapFile(
        Sys* pSys,
        const char* pcContext,
        const char* pcFile,
        MAPPING eMapping,
        uint64_t gOffset,
        uint32_t* pqRequiredSize,
        void** ppxMap
        )
{
    int iStatRet;
    int iFlags;
    int iFile;
    int iMode;
    int iProt;
    int iMap;
    int errnosave;
    struct stat tStat;

    iStatRet = stat(pcFile,&tStat);
    errnosave = errno;
    switch(eMapping) {
    default:                /* FallThrough */
    case MAPPING_R_FIXED:
        iFlags = O_RDONLY | O_SYNC;
        if(-1 == iStatRet) {
            Warn(0,errnosave,"%s '%s'",pcContext,pcFile);
            return false;
        } else if(!S_ISREG(tStat.st_mode)) {
            Warn(0,errnosave,"%s '%s' is not a regular file",pcContext,pcFile);
            return false;
        } else if(tStat.st_size != *pqRequiredSize) {
            Warn(0,0,"%s '%s': Expected size %u, got %u",
                    pcContext,pcFile,*pqRequiredSize,(uint32_t)tStat.st_size);
            return false;
        }
        iFile = open(pcFile,iFlags);
        iProt = PROT_READ;
        iMap = MAP_PRIVATE;
        break;

    case MAPPING_UIO:                   /* FallThrough */
    case MAPPING_W_FIXED:               /* FallThrough */
    case MAPPING_W_FIXED_MAY_CREATE:
        iFlags = O_RDWR | O_SYNC;
        if(-1 == iStatRet) {
            if(eMapping == MAPPING_W_FIXED_MAY_CREATE) {
                iFlags |= O_CREAT;
            } else {
                Die(pSys,errnosave,"%s",pcContext);
                /* Not reached */
                return false;
            }
        } else if((eMapping == MAPPING_UIO) && !S_ISCHR(tStat.st_mode)) {
            Warn(0,errnosave,"%s '%s' is not a character device",pcContext,pcFile);
            return false;
        } else if((eMapping != MAPPING_UIO) && !S_ISREG(tStat.st_mode)) {
            Warn(0,errnosave,"%s '%s' is not a regular file",pcContext,pcFile);
            return false;
        } else if((eMapping == MAPPING_W_FIXED_MAY_CREATE) && (tStat.st_size != *pqRequiredSize)) {
            Warn(0,0,"%s '%s': Expected size %u, got %u",
                    pcContext,pcFile,*pqRequiredSize,(uint32_t)tStat.st_size);
            return false;
        }
        iMode = 0
            | S_IRUSR
            | S_IWUSR
            | S_IRGRP
            | S_IROTH
            ;
        iFile = open(pcFile,iFlags,iMode);
        iProt = PROT_READ | PROT_WRITE;
        iMap = MAP_SHARED;
        break;
    }
    if(iFile < 0) {
        Warn(pSys,errno,"open(%s '%s')",pcContext,pcFile);
        return false;
    }
    if(iFlags & O_CREAT) {
        if(0 != ftruncate(iFile,*pqRequiredSize)) {
            Warn(pSys,errno,"ftruncate(%s '%s')",pcContext,pcFile);
            (void)close(iFile);
            return false;
        }
        if(0 != close(iFile)) {
            Warn(pSys,errno,"close(%s '%s')",pcContext,pcFile);
            return false;
        }
        iFlags &= ~O_CREAT;
        iFile = open(pcFile,iFlags,iMode);
        if(iFile < 0) {
            Warn(pSys,errno,"open(%s '%s')",pcContext,pcFile);
            return false;
        }
        iStatRet = stat(pcFile,&tStat);
        errnosave = errno;
        if(-1 == iStatRet) {
            Die(pSys,errnosave,"%s",pcContext);
            /* Not reached */
            return false;
        }
    }

    if(eMapping != MAPPING_UIO) {
        *pqRequiredSize = (uint32_t)tStat.st_size;
    }

    *ppxMap = mmap(
                0,                  /* addr */
                *pqRequiredSize,    /* length */
                iProt,              /* prot */
                iMap,               /* flags */
                iFile,              /* fd */
                (off_t)gOffset      /* offset */
                );
    errnosave = errno;

    if(0 != close(iFile)) {
        Warn(pSys,errno,"close(%s '%s')",pcContext,pcFile);
        return false;
    }

    if(MAP_FAILED == *ppxMap) {
        Warn(pSys,errnosave,"mmap(%s '%s')",pcContext,pcFile);
        return false;
    }

    return true;
}

bool UnmapFile(Sys* pSys,uint32_t qSize,void* pxMap)
{
    int iRet;

    iRet = munmap(pxMap,qSize);
    if(0 != iRet) {
        Warn(pSys,errno,"munmap(%p,0x%08x)",pxMap,qSize);
        return false;
    }
    return true;
}
